@@ -43,8 +43,9 @@ sub new {
   my $self = $class->SUPER::new($mailsaobject);
   bless ($self, $class);
 
-  $self->register_eval_rule("have_any_bounce_relays");
-  $self->register_eval_rule("check_whitelist_bounce_relays");
+  $self->register_eval_rule("have_any_bounce_relays"); # type does not matter
+  $self->register_eval_rule("check_welcomelist_bounce_relays"); # type does not matter
+  $self->register_eval_rule("check_whitelist_bounce_relays"); # type does not matter - #Stub - Remove in SA 4.1
 
   $self->set_config($mailsaobject->{conf});
 
@@ -63,7 +64,9 @@ SpamAssassin handles incoming email messages.
 
 =over 4
 
-=item whitelist_bounce_relays hostname [hostname2 ...]
+=item welcomelist_bounce_relays hostname [hostname2 ...]
+
+Previously whitelist_bounce_relays which will work interchangeably until 4.1.
 
 This is used to 'rescue' legitimate bounce messages that were generated in
 response to mail you really *did* send. List the MTA relay hostnames that
@@ -76,14 +79,15 @@ Specifically, C<*> and C<?> are allowed, but all other metacharacters are not.
 Regular expressions are not used for security reasons.
 
 Multiple addresses per line, separated by spaces, is OK.  Multiple
-C<whitelist_bounce_relays> lines are also OK.
+C<welcomelist_bounce_relays> lines are also OK.
 
 =back
 
 =cut
 
   push (@cmds, {
-      setting => 'whitelist_bounce_relays',
+      setting => 'welcomelist_bounce_relays',
+      aliases => ['whitelist_bounce_relays'], # backward compatible - to be removed for 4.1
       type => $Mail::SpamAssassin::Conf::CONF_TYPE_ADDRLIST
     });
 
@@ -92,11 +96,11 @@ C<whitelist_bounce_relays> lines are also OK.
 
 sub have_any_bounce_relays {
   my ($self, $pms) = @_;
-  return $pms->{conf}->{whitelist_bounce_relays} &&
-         %{$pms->{conf}->{whitelist_bounce_relays}} ? 1 : 0;
+  return $pms->{conf}->{welcomelist_bounce_relays} &&
+         %{$pms->{conf}->{welcomelist_bounce_relays}} ? 1 : 0;
 }
 
-sub check_whitelist_bounce_relays {
+sub check_welcomelist_bounce_relays {
   my ($self, $pms) = @_;
 
   return 0  if !$self->have_any_bounce_relays($pms);
@@ -111,7 +115,7 @@ sub check_whitelist_bounce_relays {
   foreach my $line (@{$body}) {
     next unless ($line =~ /^[> ]*Received:/i);
     while ($line =~ / (\S+\.\S+) /g) {
-      return 1 if $self->_relay_is_in_whitelist_bounce_relays($pms, $1);
+      return 1 if $self->_relay_is_in_welcomelist_bounce_relays($pms, $1);
     }
   }
 
@@ -143,18 +147,19 @@ sub check_whitelist_bounce_relays {
 
     next unless ($fullhdr =~ /^[> ]*Received:/i);
     while ($fullhdr =~ /\s(\S+\.\S+)\s/gs) {
-      return 1 if $self->_relay_is_in_whitelist_bounce_relays($pms, $1);
+      return 1 if $self->_relay_is_in_welcomelist_bounce_relays($pms, $1);
     }
   }
 
   return 0;
 }
+*check_whitelist_bounce_relays = \&check_welcomelist_bounce_relays; # removed in 4.1
 
-sub _relay_is_in_whitelist_bounce_relays {
+sub _relay_is_in_welcomelist_bounce_relays {
   my ($self, $pms, $relay) = @_;
   return 1 if $self->_relay_is_in_list(
-        $pms->{conf}->{whitelist_bounce_relays}, $pms, $relay);
-  dbg("rules: relay $relay doesn't match any whitelist");
+        $pms->{conf}->{welcomelist_bounce_relays}, $pms, $relay);
+  dbg("rules: relay $relay doesn't match any welcomelist");
 
   return 0;
 }
@@ -167,7 +172,7 @@ sub _relay_is_in_list {
   if (defined $list->{$relay}) { return 1; }
 
   foreach my $regexp (values %{$list}) {
-    if ($relay =~ qr/$regexp/i) {
+    if ($relay =~ $regexp) {
       dbg("rules: relay $relay matches regexp: $regexp");
       return 1;
     }
